/* This file is part of nanolisp, a very small Lisp system.

Copyright (C) 2007  Luca Saiu

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#ifndef NANOLISP_H_
#define NANOLISP_H_

/* We want the bool type, defined in the newer C99 standard and not included
   by default */
#include <stdbool.h>

/* I use longjmp() to be able to come back to the REPL from everywhere in the
   code after an error: */
#include <setjmp.h>

/* A nanolisp s-expression is a pointer to an opaque struct. The user
   should not make assumptions on its content: */
typedef struct nanolisp_sexpression_structure_struct*
  nanolisp_sexpression_t;

/* This is a C enum containing the type of a nanolisp s-expression; for
   example an s-expression can be an integer, a symbol, a cons, ...: */
typedef enum{
  nanolisp_integer_type,
  nanolisp_symbol_type,
  nanolisp_cons_type,
  nanolisp_boolean_type,
  nanolisp_empty_list_type,
  nanolisp_primitive_procedure_type,
  nanolisp_procedure_type
} nanolisp_type_t;

/* nanolisp_primitive_procedure_t is just a shorter name for the
   pointer-to-function type used to implement all nanolisp primitive
   procedures. A primitive procedure is a C function implementing a
   predefined nanolisp procedure: the C function takes some unspecified
   number of parameters, all of type nanolisp_sexpression_t, and returns
   a single nanolisp value as result.
   Yes, I know, C has a very ugly syntax for pointer-to-function
   types: this is the definition of a type called
   nanolisp_primitive_procedure_t. */
typedef nanolisp_sexpression_t (*nanolisp_primitive_procedure_t)();

/* Check that the given s-expression has the given type, and fail by calling
   nanolisp_error() if this is not the case: */
void nanolisp_assert_type(nanolisp_sexpression_t a, nanolisp_type_t type);

/* Given the C string 'text' return a nanolisp symbol holding it (see
   nanolisp_sexpression_structure_struct_t for information about how a
   symbol is internally implemented). If this is the first time we
   see 'text' then a new symbol is created (this is called "symbol
   interning"); otherwise an existing symbol is returned.
   This makes comparation of symbols very efficient: two symbols can always
   be compared by identity (checking whether their pointers are equal). */
nanolisp_sexpression_t nanolisp_make_symbol(char *text);

/* Return a new nanolisp object containing the given C datum:
   The empty list, true and false are created once and for all, and are
   implemented as three unique objects; in this case the nanolisp_make_XXX()
   functions have just the purpose of making the API uniform. */
nanolisp_sexpression_t nanolisp_make_integer(int value);
nanolisp_sexpression_t nanolisp_make_cons(nanolisp_sexpression_t a,
                                          nanolisp_sexpression_t b);
nanolisp_sexpression_t nanolisp_make_procedure(nanolisp_sexpression_t parameters,
                                               nanolisp_sexpression_t code,
                                               nanolisp_sexpression_t environment);
nanolisp_sexpression_t nanolisp_make_primitive_procedure(
                              int arity, // the number of parameters
                              nanolisp_primitive_procedure_t value);
nanolisp_sexpression_t nanolisp_make_boolean(bool value);
nanolisp_sexpression_t nanolisp_make_empty_list(void);

/* Return the type of the given s-expression: */
nanolisp_type_t nanolisp_get_type(nanolisp_sexpression_t a);

/* Return the content of the given s-expression: */
union nanolisp_expression_content* nanolisp_get_content(nanolisp_sexpression_t a);

/* This is the internal implementation of a nanolisp cons. As you may guess
   it's very simple: it's just a struct with two fields, the 'car' and the
   'cdr'. Note that the car and the cdr may be *any* s-expression, and in
   particular they may also be conses: */
typedef struct{
  /* The 'car' is simply the first part of a cons: */
  nanolisp_sexpression_t car;
  /* The 'cdr' is simply the second part of a cons: */
  nanolisp_sexpression_t cdr;
} nanolisp_cons_t;

/* A nanolisp non-native procedure (this means: a procedure implemented in
   nanolisp, not a procedure already implemented in C) is implemented with
   three fields:
   - a nanolisp list of symbols which are the names of the parameters
   - the code for the procedure: note that in Lisp also the code is an
     s-expression!
   - the environment, which is an nanolisp 'association list' (a list of
     conses) binding each non-local variable (as a symbol) to its value.
     For example if in the environment the variable 'x' is bound to the
     integer '2' and the variable 'y' is bound to the symbol 'foo' then
     the environment is ((x . 2) (y . foo)) */
typedef struct{
  /* A nanolisp lisp containing the formal parameter names; for example
     (x y) for a procedure which has two parameters called 'x' and 'y': */
  nanolisp_sexpression_t parameters;
  /* The code for the procedure; for example (/ (+ x y) 2) for a function
     computing the average of x and y: */
  nanolisp_sexpression_t code;
  /* The environment for the procedure, implemented as an association list:
     for example ((x . 42)) for the nanolisp procedure
     ((lambda (x)
         (lambda (y)
            (+ x y)))
      42) */
  nanolisp_sexpression_t environment;
} nanolisp_procedure_t;

/* A nanolisp primitive procedure contains an integer, representing the
   number of parameters which the procedure takes, and a pointer to the
   (C) function implementing it: */
typedef struct{
  /* The number of parameters: */
  int arity;
  /* The actual pointer-to-function: */
  nanolisp_primitive_procedure_t primitive_procedure;
} nanolisp_primitive_procedure_structure_t;

/* This is the internal implementation of each nanolisp s-expression:
   it's a two-field struct, holding the 'type' (for example
   nanolisp_integer_type, if the s-expression is an integer) and the
   'content' (for example the number 42): */
typedef struct nanolisp_sexpression_structure_struct{
  /* 'type' is an enum value representing the type of this particular
     s-expression: for example it will be nanolisp_integer_type if this
     s-expression is an integer, and nanolisp_boolean_type if this
     s-expression is a boolean. */
  nanolisp_type_t type;
  /* The field 'content' is the actual content of this s-expression.
     'content' is a C union, hence only *one* field among the following
     ones (integer_content, symbol_content, etc.) can be used at any
     given time: the value of 'type' determines which field is used, if
     any. */
  union nanolisp_expression_content{
    /* 'integer_content' is the integer stored in the sexpression: */
    int integer_content;
    /* 'symbol_content' is the index of the symbol stored in the
       sexpression. It would not be really needed, because symbols
       are also uniquely identified by their *pointers*. This could be
       changed. */
    int symbol_content;
    /* 'cons_content' is simply a struct containing the car and the cdr. */
    nanolisp_cons_t cons_content;
    /* 'boolean_content' is either true or false: */
    bool boolean_content;
    /* 'primitive_procedure_content' is a structure containing a pointer to
       the C function which implements a nanolisp primitive function, and its
       arity (number of paramters): */
    nanolisp_primitive_procedure_structure_t primitive_procedure_content;
    /* 'procedure_content' is simply a nanolisp_procedure_t containing
       parameters names, code and environment for a nanolisp procedure
       (a nanolisp "lambda") */
    nanolisp_procedure_t procedure_content;
  } content;
} nanolisp_sexpression_structure_struct_t;

/* Print the given error message, then exit with failure by longjmp()ing
   to an appropriate location: */
void nanolisp_error(const char *error_message)
  __attribute__(( noreturn ));

/* Return a C boolean which is true if and only if the given
   s-expression a has the given type; for example nanolisp_is_cons()
   returns a C true value if its parameter is a nanolisp cons;
   if it's any other nanolisp s-expression then return the C
   boolean false: */
bool nanolisp_is_integer(nanolisp_sexpression_t a);
bool nanolisp_is_symbol(nanolisp_sexpression_t a);
bool nanolisp_is_cons(nanolisp_sexpression_t a);
bool nanolisp_is_boolean(nanolisp_sexpression_t a);
bool nanolisp_is_empty_list(nanolisp_sexpression_t a);
bool nanolisp_is_primitive_procedure(nanolisp_sexpression_t a);
bool nanolisp_is_procedure(nanolisp_sexpression_t a);

/* Return a C boolean telling whether a is a proper (i.e. non-dotted) list:
   by definition either:
   i) an empty list
   or:
   ii) a cons whose cdr is a proper list: */
bool nanolisp_is_list(nanolisp_sexpression_t a);

/* Extract the content from an s-expression and return it as a C value
   (for example, if the s-expression is a nanolisp boolean, then return
   a C bool; if the s-expression is a nanolisp cons, then return a C
   pointer to a nanolisp_cons_t).
   Fail by calling nanolisp_error() if the type is wrong (for example
   if nanolisp_get_symbol() is called on a nanolisp boolean). */
int nanolisp_get_integer(nanolisp_sexpression_t a);
int nanolisp_get_symbol_index(nanolisp_sexpression_t a);
char* nanolisp_get_symbol_as_c_string(nanolisp_sexpression_t a);
nanolisp_cons_t* nanolisp_get_cons(nanolisp_sexpression_t a);
bool nanolisp_get_boolean(nanolisp_sexpression_t a);
nanolisp_primitive_procedure_structure_t*
   nanolisp_get_primitive_procedure(nanolisp_sexpression_t a);
nanolisp_procedure_t* nanolisp_get_procedure(nanolisp_sexpression_t a);

/* Perform arithmetics on two nanolisp objects. Fail with nanolisp_error()
   if at least one of the parameter is not an integer, or in case of
   division or remainder by zero. Remember that the C operator for
   remainder is '%'.
   Remember that nanolisp_get_integer() already does the type checking for
   you and calls nanolisp_error() when appropriate for type errors.
   Of course, as the type says, the result must be a nanolisp s-expression,
   and not a C int. */
nanolisp_sexpression_t nanolisp_plus(nanolisp_sexpression_t a,
                                     nanolisp_sexpression_t b);
nanolisp_sexpression_t nanolisp_minus(nanolisp_sexpression_t a,
                                      nanolisp_sexpression_t b);
nanolisp_sexpression_t nanolisp_times(nanolisp_sexpression_t a,
                                      nanolisp_sexpression_t b);
nanolisp_sexpression_t nanolisp_divided(nanolisp_sexpression_t a,
                                        nanolisp_sexpression_t b);
nanolisp_sexpression_t nanolisp_remainder(nanolisp_sexpression_t a,
                                          nanolisp_sexpression_t b);

/* Perform comparations on two nanolisp objects which must be integers.
   These are very similar to the arithmetic operators above, but return
   nanolisp booleans instead of nanolisp integers: */
nanolisp_sexpression_t nanolisp_less(nanolisp_sexpression_t a,
                                     nanolisp_sexpression_t b);
nanolisp_sexpression_t nanolisp_less_or_equal(nanolisp_sexpression_t a,
                                              nanolisp_sexpression_t b);
nanolisp_sexpression_t nanolisp_greater(nanolisp_sexpression_t a,
                                        nanolisp_sexpression_t b);
nanolisp_sexpression_t nanolisp_greater_or_equal(nanolisp_sexpression_t a,
                                                 nanolisp_sexpression_t b);

/* Return the car of the cdr of the given s-expression 'a'. Fail with
   nanolisp_error() if 'a' is not a cons.
   Remember that nanolisp_get_cons() already does the type checking for
   you and calls nanolisp_error() when appropriate for type errors. */
nanolisp_sexpression_t nanolisp_car(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_cdr(nanolisp_sexpression_t a);

/* Return a nanolisp true s-expression if the two given s-expressions
   'a' and 'b' are equal, otherwise return the nanolisp false s-expression.
   This function *never* fails: if the types of 'a' and 'b' are different
   it just returns false. Structural equality (*not* equality-by-identity)
   is used*/
nanolisp_sexpression_t nanolisp_equal(nanolisp_sexpression_t a,
                                      nanolisp_sexpression_t b);

/* The 'car' of the 'car' can be also called 'caar'. The 'cdr' of the 'cdr'
   of the 'car' can also be called 'cddar'. 
   This is very easy if you have implemented 'car' and 'cdr' in the right
   way. */
nanolisp_sexpression_t nanolisp_caar(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_cadr(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_cdar(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_cddr(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_caaar(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_caadr(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_cadar(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_caddr(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_cdaar(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_cdadr(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_cddar(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_cdddr(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_caaaar(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_caaadr(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_caadar(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_caaddr(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_cadaar(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_cadadr(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_caddar(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_cadddr(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_cdaaar(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_cdaadr(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_cdadar(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_cdaddr(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_cddaar(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_cddadr(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_cdddar(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_cddddr(nanolisp_sexpression_t a);

/* Utility procedures: for example make_list2(X, Y) is equivalent to
   make_cons(X, make_cons(Y, empty_list)): */
nanolisp_sexpression_t nanolisp_make_list1(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_make_list2(nanolisp_sexpression_t a,
                                           nanolisp_sexpression_t b);
nanolisp_sexpression_t nanolisp_make_list3(nanolisp_sexpression_t a,
                                           nanolisp_sexpression_t b,
                                           nanolisp_sexpression_t c);
nanolisp_sexpression_t nanolisp_make_list4(nanolisp_sexpression_t a,
                                           nanolisp_sexpression_t b,
                                           nanolisp_sexpression_t c,
                                           nanolisp_sexpression_t d);

/* Give the name 'name' to the value 'value', by updating the
   global environment, and always return the nanolisp true value.
   If name is not a symbol then fail with nanolisp_error(). 'value'
   can be a nanolisp s-expression of any type. */
nanolisp_sexpression_t nanolisp_define(nanolisp_sexpression_t name,
                                       nanolisp_sexpression_t value);

/* Parse the given string (which should contain a sequence of s-expressions)
   and, in case of success, write the result (as a nanolisp list) into the
   given pointer.
   Return 0 in case of success, and a nonzero value in case of errors. */
int nanolisp_parse_string(const char *text, nanolisp_sexpression_t *result);

/* Parse the content of the given FILE* (which should contain a sequence of
   s-expressions) and, in case of success, write the result (as a nanolisp
   list) into the given pointer.
   Return 0 in case of success, and a nonzero value in case of errors. */
int nanolisp_parse_FILE_star(FILE *file, nanolisp_sexpression_t *result);

/* Parse the given C string into a new nanolisp s-expression, and
   return it. Fail with nanolisp_error() if there is any error. */
nanolisp_sexpression_t nanolisp_read(char *string);

/* Read a single s-expression in a single line from the terminal; return
   it or fail on any error: */
nanolisp_sexpression_t nanolisp_read_from_terminal(void);

/* Print to the stardard output the given s-expression 'a'. Hint:
   you should use recursion in the C implementation; it's MUCH more
   difficult if you don't do so. */
void nanolisp_print(nanolisp_sexpression_t a);

/* Enter the Read-Eval-Print Loop, or "REPL".
   The REPL is an infinite loop in which, for each iteration:
   - the system *reads* a string from the terminal and parses it into
     a nanolisp s-expression
   - the system *evaluates* the s-expression (i.e. interprets it)
   - the system *prints* the result to the terminal.
   Then the next iteration of the loop starts. */
void nanolisp_repl(void);

/* Return the value which is currently bound to the given symbol in the global
   environment, or NULL if the given symbol is unbound. Fail if the given object
   is not a symbol: */
nanolisp_sexpression_t nanolisp_lookup_global(nanolisp_sexpression_t symbol);

/* Bind the given value to the given symbol in the global environment; the
   previous binding, if any, is lost. Fail if the given object
   is not a symbol: */
void nanolisp_set_global(nanolisp_sexpression_t symbol,
                         nanolisp_sexpression_t value);


void initialize_nanolisp(void);
void initialize_primitives(void);

extern jmp_buf nanolisp_where_to_jump_in_case_of_error;

/* Write the given s-expression to the terminal, and return the nanolisp true
   object: */
nanolisp_sexpression_t nanolisp_display(nanolisp_sexpression_t a);

/* Write a newline to the terminal, and return the nanolisp true object: */
nanolisp_sexpression_t nanolisp_newline(void);

/* Exit from nanolisp in a clean way: */
nanolisp_sexpression_t nanolisp_quit(void)
  __attribute__ (( noreturn ));

/* These are equivalent to nanolisp_is_XXXX(a), but return a nanolisp boolean instead
   of a C boolean; hence they are suitable to be used as primitives.
   The only notable difference is in nanolisp_procedure_p(), which does not distinguish
   between primitive and non-primitive procedures, as the difference does not matter to
   the user: */
nanolisp_sexpression_t nanolisp_integer_p(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_number_p(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_symbol_p(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_boolean_p(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_cons_p(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_null_p(nanolisp_sexpression_t a);
nanolisp_sexpression_t nanolisp_procedure_p(nanolisp_sexpression_t a); // primitive or non-primitive

/* Return the nanolisp true value iff the parameter is 0: */
nanolisp_sexpression_t nanolisp_zero_p(nanolisp_sexpression_t a);

/* Print licensing information and return the nanolisp true object: */
nanolisp_sexpression_t nanolisp_print_no_warranty(void);
nanolisp_sexpression_t nanolisp_print_license(void);

/* Perform a garbage collection and return the nanolisp true object; this is
   only useful for debugging: of course the garbage collector is automatically
   invoked: */
nanolisp_sexpression_t nanolisp_gc(void);

#endif // #ifndef NANOLISP_H_
